@@ -8,13 +8,25 @@ from urllib.request import urlopen
 
 class Gutenberg_Dataset():
     """ A fuzzy, lightweight class to access, search and filter Project Gutenberg resources
-        
-        GutenbergLib by default uses a mirror's root URL. :func:`~Gutenberg_Dataset.Gutenberg_Dataset.load_index` 
-        needs to be called before any other methods.
-        
-        :param root_url: url of Project Gutenberg or any mirror URL.
-        :param cache_dir: path to a directory that will be used to cache the Gutenberg index and already downloaded texts
-        """
+    
+    GutenbergLib by default uses a mirror's root URL. Alternatively, you can specify a local directory containing a Gutenberg mirror.
+    That mirror directory needs to contain a GUTINDEX.ALL file and has typically many 
+    sub-directories `0` ,.. `n` .
+
+    A mirror of project Gutenberg can be created by:
+    
+    .. code-block:: console
+
+        #!/bin/bash
+        rsync -zarv --dry-run --prune-empty-dirs --del --include="*/" --include='*.'{txt,pdf,ALL} --exclude="*" aleph.gutenberg.org::gutenberg ./gutenberg_mirror
+
+    You can remove the PDF files, since they are currently not used, and need to review the `--dry-run` option.
+
+    Note: :func:`~Gutenberg_Dataset.Gutenberg_Dataset.load_index` needs to be called before any other methods.
+    
+    :param root_url: url of Project Gutenberg or any mirror URL, or a local directory containing a Gutenberg mirror.
+    :param cache_dir: path to a directory that will be used to cache the Gutenberg index and already downloaded texts. The cache directory is only used, if a remote Gutenberg URL and not a local mirror is used.
+    """
     def __init__(self, root_url="http://www.mirrorservice.org/sites/ftp.ibiblio.org/pub/docs/books/gutenberg", cache_dir="gutenberg"):
         self.log = logging.getLogger('GutenbergLib')
         self.root_url = root_url
@@ -24,18 +36,37 @@ class Gutenberg_Dataset():
                            "This book was generously provided by the ", "*** START OF THIS PROJECT GUTENBERG",
                            "START OF THE PROJECT GUTENBERG"]    
         self.near_start_tokens=["produced by ", "Produced by ", ", prepared by", "Transcriber's Note", 
-                                "Transcriber's note:", "Anmerkungen zur Tanskription", "Distributed Proofreading Team"] 
+                                "Transcriber's note:", "Anmerkungen zur Tanskription", 
+                                "Distributed Proofreading Team", "offensichtliche Schreibfehler",
+                                "Inkonsistenzen in der Rechtschreibung", "Im Original",
+                                "Obvious printer errors", "spelling was kept",
+                                "ERRATA have been applied", "punctuation errors", "have been silently corrected",
+                                "changes to the text", "Transcriber note", "Transcriber Note","_italic_",
+                                "Variable spelling"] 
         self.end_tokens=["End of the Project Gutenberg", "*** END OF THIS PROJECT", 
                          "***END OF THE PROJECT GUTENBER", "Ende dieses Projekt Gutenberg", 
                          "*** END OF THE PROJECT GUTENBERG",
                          "End of Project Gutenberg", "Transcriber's Note"]
-        try:
-            if not os.path.exists(cache_dir):
-                os.makedirs(cache_dir)
-            self.cache_dir=cache_dir
-        except Exception as e:
-            self.cache_dir=None
-            self.log.error(f"Failed to create cache directory {cache_dir}, {e}")
+        self.local_mirror=False
+        if root_url[:4]!='http':
+            if not os.path.exists(root_url):
+                self.log.error(f'If root_url points to non-http URL, it must be an existing local directory containing a Gutenberg mirror: {root_url}')
+            else:
+                index_path=os.path.join(root_url, "GUTINDEX.ALL")
+                if not os.path.exists(index_path):
+                    self.log.error(f'GUTINDEX.ALL not found in {root_url}, this is not a valid Gutenberg mirror')
+                else:
+                    self.local_mirror=True
+                    self.cache_dir=None
+            return
+        if self.local_mirror is False:
+            try:
+                if not os.path.exists(cache_dir):
+                    os.makedirs(cache_dir)
+                self.cache_dir=cache_dir
+            except Exception as e:
+                self.cache_dir=None
+                self.log.error(f"Failed to create cache directory {cache_dir}, {e}")
 
     def _parse_record(self,record,verbose=True):
         """ internal function to recreate some consistent record information from near-freestyle text """
@@ -222,57 +253,70 @@ class Gutenberg_Dataset():
         :param cache_expire_days: Number of days after which the index is re-downloaded.
         """
         raw_index=None
-        if self.cache_dir is None:
-            self.log.error("Cannot cache library index, no valid cache directory.")
-            return False
-        ts_file=os.path.join(self.cache_dir,"timestamp")
-        cache_file=os.path.join(self.cache_dir,"gutenberg_index")
-        expired=True
-        read_from_cache=False
-        if os.path.isfile(ts_file) and os.path.isfile(cache_file):
-            try:
-                with open(ts_file,'r') as f:
-                    ts=float(f.read())
-                if time.time()-ts<cache_expire_days*24*3600:
-                    expired=False
-                    read_from_cache = True
-                    self.log.debug("Cache timestamp read.")
-                else:
-                    self.log.info("Cache for Gutenberg-index is expired, reloading from web.")
-            except:
-                self.log.warning("Failed to read cache timestamp, reloading from web.")
-        if expired is False and os.path.isfile(cache_file):
-            try:
-                with open(cache_file,'r') as f:
-                    raw_index=f.read()
-                    self.log.info(f"Gutenberg index read from local cache: {cache_file}")
-            except:
-                expired=True
-                self.log.warning("Failed to read cached index, reloading from web.")
-        if expired is True:
-            index_url=self.root_url+"/GUTINDEX.ALL"
-            try:
-                raw_index = urlopen(index_url).read().decode('utf-8')
-                if raw_index[0]=='\ufeff':  # Ignore BOM
-                    raw_index=raw_index[1:]
-                raw_index=raw_index.replace('\r','')
-                self.log.info(f"Gutenberg index read from {index_url}")
-            except Exception as e:
-                self.log.error(f"Failed to download Gutenberg index from {index_url}, {e}")
+        if self.local_mirror is False:
+            if self.cache_dir is None:
+                self.log.error("Cannot cache library index, no valid cache directory.")
                 return False
-        if cache is True and read_from_cache is False:
+            ts_file=os.path.join(self.cache_dir,"timestamp")
+            cache_file=os.path.join(self.cache_dir,"gutenberg_index")
+            expired=True
+            read_from_cache=False
+            if os.path.isfile(ts_file) and os.path.isfile(cache_file):
+                try:
+                    with open(ts_file,'r') as f:
+                        ts=float(f.read())
+                    if time.time()-ts<cache_expire_days*24*3600:
+                        expired=False
+                        read_from_cache = True
+                        self.log.debug("Cache timestamp read.")
+                    else:
+                        self.log.info("Cache for Gutenberg-index is expired, reloading from web.")
+                except:
+                    self.log.warning("Failed to read cache timestamp, reloading from web.")
+            if expired is False and os.path.isfile(cache_file):
+                try:
+                    with open(cache_file,'r') as f:
+                        raw_index=f.read()
+                        self.log.info(f"Gutenberg index read from local cache: {cache_file}")
+                except:
+                    expired=True
+                    self.log.warning("Failed to read cached index, reloading from web.")
+            if expired is True:
+                index_url=self.root_url+"/GUTINDEX.ALL"
+                try:
+                    raw_index = urlopen(index_url).read().decode('utf-8')
+                    if raw_index[0]=='\ufeff':  # Ignore BOM
+                        raw_index=raw_index[1:]
+                    raw_index=raw_index.replace('\r','')
+                    self.log.info(f"Gutenberg index read from {index_url}")
+                except Exception as e:
+                    self.log.error(f"Failed to download Gutenberg index from {index_url}, {e}")
+                    return False
+            if cache is True and read_from_cache is False:
+                try:
+                    with open(ts_file,'w') as f:
+                        f.write(str(time.time()))
+                        self.log.debug("Wrote read cache timestamp.")
+                except Exception as e:
+                    self.log.error(f"Failed to write cache timestamp to {ts_file}, {e}")
+                try:
+                    with open(cache_file,'w') as f:
+                        f.write(raw_index)
+                        self.log.debug("Wrote read cached index.")
+                except Exception as e:
+                    self.log.error(f"Failed to write cached index to {cache_file}, {e}")
+        else:
+            index_file=os.path.join(self.root_url, "GUTINDEX.ALL")
             try:
-                with open(ts_file,'w') as f:
-                    f.write(str(time.time()))
-                    self.log.debug("Wrote read cache timestamp.")
+                with open(index_file,'r') as f:
+                    raw_index=f.read()
+                    if raw_index[0]=='\ufeff':  # Ignore BOM
+                        raw_index=raw_index[1:]
+                    raw_index=raw_index.replace('\r','')
+                    self.log.info(f"Gutenberg index read from local mirror: {index_file}")
             except Exception as e:
-                self.log.error(f"Failed to write cache timestamp to {ts_file}, {e}")
-            try:
-                with open(cache_file,'w') as f:
-                    f.write(raw_index)
-                    self.log.debug("Wrote read cached index.")
-            except Exception as e:
-                self.log.error(f"Failed to write cached index to {cache_file}, {e}")
+                self.log.error(f"Failed to read Gutenberg index from local mirror: {index_file}, {e}")
+                return
         lines=raw_index.split('\n')
         self.records=self._parse_index(lines)
 
@@ -285,7 +329,11 @@ class Gutenberg_Dataset():
         :param ebook_id: Gutenberg id (Note: string, since this sometimes contains a character!)
         :returns: book text as string, unfiltered. Can be filtered with :func:`~Gutenberg_Dataset.Gutenberg_Dataset.filter_text`
         """
-        return self._load_book_ex(ebook_id)[0]
+        txt, dl, val = self._load_book_ex(ebook_id)[0]
+        if val is True:
+            return txt, dl
+        else:
+            return None, dl
 
     def _read_download(self, filenames, path_stub, cache_name):
         """ Internal function to read ebook from cache or download it. """
@@ -306,16 +354,30 @@ class Gutenberg_Dataset():
         file_url=None
         for filename, encoding in filenames:
             file_url=self.root_url+path_stub+filename
-            try:
-                if encoding != 'bin':
-                    data = urlopen(file_url).read().decode(encoding)
-                else:
-                    data = urlopen(file_url).read()
-                self.log.info(f"Book downloaded from {file_url}")
-                downloaded = True
-                break
-            except Exception as e:
-                self.log.debug(f"URL-Download failed: {file_url}, {e}")
+            if self.local_mirror is False:
+                try:
+                    if encoding != 'bin':
+                        data = urlopen(file_url).read().decode(encoding)
+                    else:
+                        data = urlopen(file_url).read()
+                    self.log.info(f"Book read from {file_url}")
+                    downloaded = True
+                    break
+                except Exception as e:
+                    self.log.debug(f"Failed to download {file_url}, {e}")
+            else:
+                try:
+                    if encoding != 'bin':
+                        with open(file_url,'r',encoding=encoding) as f:
+                            data = f.read()
+                    else:
+                        with open(file_url,'rb') as f:
+                            data = f.read()
+                    self.log.info(f"Book read from local mirror at {file_url}")
+                    downloaded = False
+                    break
+                except Exception as e:
+                    self.log.debug(f"Failed to read local mirror at {file_url}, {e}")
         return data, cache_file, downloaded
 
     def _load_book_ex(self, ebook_id):
@@ -330,7 +392,7 @@ class Gutenberg_Dataset():
         Use :func:`~Gutenberg_Dataset.Gutenberg_Dataset.get_book` to retrieve a dictionary with metadata and filtered text.
 
         :param ebook_id: Gutenberg id (Note: string, since this sometimes contains a character!)
-        :returns: tuple: book text as string, unfiltered, and a flag indicating with 'True' if book was downloaded from remote. 
+        :returns: tuple: book text as string, unfiltered, and a flag indicating with 'True' if book was downloaded from remote, validity flag, 'True' indicates valid text. 
         """
         if ebook_id is None or len(ebook_id)==0:
             return None
@@ -338,6 +400,7 @@ class Gutenberg_Dataset():
             ebook_id=ebook_id[:-1]
         path_stub=""
         downloaded = False
+        valid = False
         
         for i in range(len(ebook_id)-1):
             path_stub+="/"+ebook_id[i]
@@ -349,6 +412,7 @@ class Gutenberg_Dataset():
             if data[0]=='\ufeff':  # Ignore BOM
                 data=data[1:]
             data=data.replace('\r','')
+            valid = True
         else:
             filenames = [(ebook_id+"-pdf.pdf","bin")]
             cache_name=ebook_id+".pdf"
@@ -357,14 +421,14 @@ class Gutenberg_Dataset():
                 self.log.error(f"Ebook {cache_name} is only available in PDF format, this is not supported.")
             else:
                 self.log.warning(f"Failed to download {filenames}, skipping book.")
-            return None, downloaded
+                return None, downloaded, False
         if cache_file is not None:
             try:
                 with open(cache_file,'w') as f:
                     f.write(data)
             except:
                 self.log.error(f"Failed to cache file {cache_file}")
-        return data, downloaded
+        return data, downloaded, valid
     
     def filter_text(self, book_text, add_start_tokens=None, add_near_start_tokens=None, add_end_tokens=None):
         """ Heuristically remove header and trailer texts not part of the actual books
@@ -534,21 +598,24 @@ class Gutenberg_Dataset():
         into the dictionary with the key `text`.
         
         :param search_dict: search array of dictionaries that at least contain the key `ebook_id`.
-        :param download_count_limit: maximum number of books to download_count_limit
+        :param download_count_limit: maximum number of books to download, if no local mirror is used. No limits apply for local mirrors.
         :returns: list of records including filtered book text-based in the `text` field.
         """
-        dlc=len(search_dict)
         dls=0
         for i in range(0, len(search_dict)):
-            bt, dl = self._load_book_ex(search_dict[i]["ebook_id"])
-            if bt is None:
-                self.log.error(f"Download of book {search_dict[i]['ebook_id']}, {search_dict[i]['title']} failed!")
+            self.log.debug(f"Getting id={search_dict[i]['ebook_id']}, {search_dict[i]['title']}")
+            bt, dl, val = self._load_book_ex(search_dict[i]["ebook_id"])
+            if bt is None or val is False:
+                if val is False:
+                    self.log.warning(f"Download of book {search_dict[i]['ebook_id']}, {search_dict[i]['title']}: invalid format!")
+                else:
+                    self.log.error(f"Download of book {search_dict[i]['ebook_id']}, {search_dict[i]['title']} failed!")
                 continue
             search_dict[i]['text']=self.filter_text(bt)
-            if dl is True:
+            if dl is True and self.local_mirror is False:
                 dls += 1
-                if dls>dlc:
-                    print(f"Download limit reached ({dlc}), stopping download...")
+                if dls>download_count_limit:
+                    self.log.error(f"Download limit reached ({download_count_limit}), stopping download...")
                     break
         return search_dict  
 
