@@ -232,6 +232,17 @@ class Text_Dataset:
             raise ValueError(f"Unknown tokenizer {tokenizer}")
         return decoded_text
 
+    def get_random_char_tokenized_sample_pair(self, length):
+        """ Get a random tokenized sample of the dataset.
+        
+        :param length: length of the sample
+        :return: tuple (X, y) encoded sample
+        """
+        _, sample = self.encode(self.get_random_sample(length+1), tokenizer='char', sanitize_white_space=False)
+        X = sample[:-1]
+        y = sample[1:]
+        return X, y
+        
     def _display_colored_html(self, textlist, dark_mode=False, display_ref_anchor=True, pre='', post=''):
         """ Internal function to display text and citation references in HTML. """
         bgcolorsWht = ['#d4e6e1', '#d8daef', '#ebdef0', '#eadbd8', '#e2d7d5', '#edebd0',
